#! /usr/bin/env python
# "@(#) $Id: AntennaPositions.py 212412 2014-12-11 20:28:17Z rmarson $"
# ALMA - Atacama Large Millimeter Array
# (c) Associated Universities Inc., 2009 - 2011
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

#************************************************************************
#   NAME StabdardInterferometry
# 
#   SYNOPSIS This script is inteded to provide basic interferometric
#            observations capabilities of single pointing targets.
#
#------------------------------------------------------------------------

import Control
import Observation.SchedulingBlock
from Observation.Global import simulatedArray,addClassToMode, OperatorLogger
from Observation.SBExecutionMode import SBExecutionMode
from CCL.Global import *
from CCL.ObservingModeBase import SBOutOfTime
from PyDataModelEnumeration import PyCalibrationDevice
from Observation.SBRatioCalTarget import SBRatioCalTarget
from CCL.logging import getLogger
import math
from Observation.TargetUtility import *


logger = OperatorLogger()

sb = Observation.SchedulingBlock.SchedulingBlock()

# Get expert parameters
numSources     = int(sb.getExpertParameter('NumSources',  default=50))
logger.logInfo("numSources set to %s " % numSources)
elLimit         = float(sb.getExpertParameter('ElLimit', default=20.))

# Get reference to the array and start the exec block
array = getArray()
array.beginExecution()

# Get the Observing Mode
obsmode = array.getInterferometryObservingMode()

# add the SB execution methods
addClassToMode(obsmode, SBExecutionMode)

# Just start on the sky, to be sure. Should be in the mode defaults.
# it is now as we use subscan sequences
# obsmode.setCalibrationDevice(PyCalibrationDevice.NONE)

# Set observing mode restrictions
logger.logInfo("Setting elevation limit to %f degrees" % elLimit)
obsmode.setElevationLimit('%f deg' % elLimit)
endTime = sb.getMaximumExecutionTime()
logger.logInfo("Setting maximum execution time to %s seconds or %s minutes" % (endTime,endTime/60.0))
obsmode.setMaximumSchedBlockTime(endTime)

# Turn off automatic WVR reduction on all targets:
if not simulatedArray():
    obsmode.disableWVRCalibration(True)
# This should be the default for real interferometry: we use reference positions only for AtmCals.
for target in sb.getTargetList():
    target.setUseReferencePosition(target in sb.getAtmCalTargetList())

# Get group list
groupList = sb.getGroupList()
#
try:
        
    # The first and only group is special because we don't expect a science target
    group = groupList[0]

    group.expandSourceQueries(group.getDelayCalTargetList(), sortQuery='FLUX')
    #
    # here we may modify the list of sources (change  order and sky distribution)
    
    obsmode.executeCalTargetList(group.getDelayCalTargetList(),
                                 minimumNumber=numSources)

    # Group is complete: remove the primaryPhaseCal from cleanup:
    logger.logInfo("Group number %s completed."% groupList.index(group))

    # All groups are complete: execute the cleanup (e.g. bandpass or amplitude cals)
    logger.logInfo("All groups are completed.")

except SBOutOfTime:
    #
    # Ok there should be just enough time for the cleanup list now
    logger.logWarning("Running short of time, forcing cleanup!")

obsmode.resetLimits()
# should be empty...
logger.logInfo("Executing cleanup list.")
obsmode.executeCleanupList()
array.endExecution(Control.SUCCESS, "Successful Completion")
           
#
# ___oOo___
