#! /usr/bin/env python
# "@(#) $Id: SingleTargetInterferometry.py 212412 2014-12-11 20:28:17Z rmarson $"
# ALMA - Atacama Large Millimeter Array
# (c) Associated Universities Inc., 2009 - 2011
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

#************************************************************************
#   NAME StabdardInterferometry
# 
#   SYNOPSIS This script is inteded to provide basic interferometric
#            observations capabilities of single pointing targets.
#
#------------------------------------------------------------------------

import Control
import Observation.SchedulingBlock
from Observation.SBExecutionMode import addSBExecutionMode
from Observation.Global import simulatedArray
from CCL.Global import *
from CCL.ObservingModeBase import SBOutOfTime
from PyDataModelEnumeration import PyCalibrationDevice
from Observation.SBRatioCalTarget import SBRatioCalTarget
from CCL.logging import getLogger
import math
from Observation.TargetUtility import *


logger = getLogger()

sb = Observation.SchedulingBlock.SchedulingBlock()

# Get expert parameters
numPhaseCal     = int(sb.getExpertParameter('NumPhaseCal',  default=1))
logger.logInfo('NumPhaseCal set to  %s' % numPhaseCal)
numAmplitudeCal = int(sb.getExpertParameter('NumAmplitudeCal', default=1))
numBandpassCal  = int(sb.getExpertParameter('NumBandpassCal', default=1))
numPointingCal  = int(sb.getExpertParameter('NumPointingCal', default=1))
numFocusCal     = int(sb.getExpertParameter('NumFocusCal', default=1))
elLimit         = float(sb.getExpertParameter('ElLimit', default=20.))
maxPointingSeparation = float(sb.getExpertParameter('MaxPointingSeparation', default=25.))

# Get reference to the array and start the exec block
array = getArray()
array.beginExecution()

# Get the Observing Mode
obsmode = array.getInterferometryObservingMode()

# add the SB execution methods
addSBExecutionMode(obsmode)

# Just start on the sky, to be sure. Should be in the mode defaults.
obsmode.setCalibrationDevice(PyCalibrationDevice.NONE)

# Set observing mode restrictions
logger.logInfo("Setting elevation limit to %f degrees" % elLimit)
obsmode.setElevationLimit('%f deg' % elLimit)
endTime = sb.getMaximumExecutionTime()
###
###endTime=600.
###
logger.logInfo("Setting maximum execution time to %s seconds or %s minutes" % (endTime,endTime/60.0))
obsmode.setMaximumSchedBlockTime(endTime)

# Turn off automatic WVR reduction on all targets:
if not simulatedArray():
    obsmode.disableWVRCalibration(True)
# This should be the default for real interferometry: we use reference positions only for AtmCals.
for target in sb.getTargetList():
    target.setUseReferencePosition(target in sb.getAtmCalTargetList())
#
# Turning off the atm cals for pointing and focus targets
#
for target in sb.getPointingCalTargetList() + sb.getFocusCalTargetList():
    target._associatedCalTarget = []
#
# Dynamically create the SBRatioCalTargets for each AtmCalTarget
#
for target in sb.getAtmCalTargetList():
    target.createAssociatedSBRatioCalTarget()
# Get group list
groupList = sb.getGroupList()
#
# Associate pointing targets with other targets
#
for group in groupList:
    group.associatePointingCalTargets(maxPointingSeparation)

try:
    # adjust Doppler once using the science targets (only the first one).
    obsmode.adjustDoppler(sb.getTargetList(), sb.getScienceTargetList())
        
    # A workaround for the bug described in COMP-5380. -> ObsTarget updated.
    #for target in sb.getTargetList():
    #    target.setSubscanDuration(target.getSubscanDuration())

    # The first group is special because we don't expect a science target
    group = groupList[0]
    obsmode.executeCalTargetList(group.getFocusCalTargetList(), minimumNumber=numFocusCal)
    numDone, numInCleanup = obsmode.executeCalTargetList(group.getBandpassCalTargetList(),
                                                          minimumNumber=numBandpassCal,
                                                          deferToCleanupList=True,
                                                          minimumInCleanupList=1)
    logger.logInfo("Bandpass Targets: %d executed, %d in Cleanup List" %(numDone, numInCleanup))
    obsmode.executeCalTargetList(group.getDelayCalTargetList())
    numDone, numInCleanup = obsmode.executeCalTargetList(group.getAmplitudeCalTargetList(),
                                            minimumNumber=numAmplitudeCal,
                                            deferToCleanupList=True)
    logger.logInfo("Amplitude Cal Targets: %d executed, %d in Cleanup List" %(numDone, numInCleanup))

    # Now the rest of the groups should contain science targets
    for group in groupList[1:]:
        # The primary phase calibrator is assumed to be the one with the
        # shortest cycle time in the group.  This effectively sets the scan
        # duration.
        primaryPhaseCal = obsmode.getPrimaryPhaseCal(group)
        logger.logInfo("The primary phase calibrator is %s "\
                       % primaryPhaseCal.getSource().sourceName)
        
        # A group is defined to be complete when all science targets within
        # the group are complete (or unobservable)
        while not group.isComplete(obsmode):
            for scienceTarget in group.getScienceTargetList():
                if scienceTarget.isObservable(obsmode):
                    obsmode.executeCalTargetList(group.getFocusCalTargetList())
                    obsmode.executeCalTargetList(group.getBandpassCalTargetList())
                    obsmode.executeCalTargetList(group.getAmplitudeCalTargetList())
                    obsmode.executeCalTargetList(group.getPhaseCalTargetList(),
                                                  minimumNumber = numPhaseCal)
                    scienceTarget.execute(obsmode, [primaryPhaseCal])
                    primaryPhaseCal.execute(obsmode)
                    if not primaryPhaseCal.isObservable(obsmode,
                                                        duration=primaryPhaseCal.getCycleTime()):
                        logger.logWarning("Attempting to find a new primary phase calibrator as %s will soon be setting"\
                                          % primaryPhaseCal.getSource().sourceName)
                        primaryPhaseCal = obsmode.getPrimaryPhaseCal(group, primaryPhaseCal)
                        logger.logWarning("The new primary phase calibrator is %s" \
                                          % primaryPhaseCal.getSource().sourceName)
                    
        # Group is complete: remove the primaryPhaseCal from cleanup:
        logger.logInfo("Group number %s completed."% groupList.index(group))
        
        obsmode.removeTargetFromCleanupList(primaryPhaseCal)

    # All groups are complete: execute the cleanup (e.g. bandpass or amplitude cals)
    logger.logInfo("All groups are completed.")

except SBOutOfTime:
    #
    # Ok there should be just enough time for the cleanup list now
    logger.logWarning("Running short of time, forcing cleanup!")

obsmode.resetLimits()
logger.logInfo("Executing cleanup list.")
obsmode.executeCleanupList()
array.endExecution(Control.SUCCESS, "Successful Completion")
           
#
# ___oOo___
