#! /usr/bin/env python
# ALMA - Atacama Large Millimeter Array
# (c) Associated Universities Inc., 2009 - 2014
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that i1t will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

#************************************************************************
#   NAME StandardInterferometry.py
# 
#   SYNOPSIS This script is intended to meet the needs of single or multi-target
#       observations which have a common spectral configuration for all
#       science targets and a shared primary phase calibrator.  
# 
#------------------------------------------------------------------------

#
# forcing global imports is due to an OSS problem
#
global math
import math
global ast
import ast
global time
from time import time

global Control
import Control

global CCL
from CCL.Global import *
from CCL.APDMSchedBlock import SpectralSpec
import CCL.APDMSchedBlock
from CCL.ObservingModeBase import SBOutOfTime

from PyDataModelEnumeration import PyCalibrationDevice

global Observation
import Observation.SchedulingBlock
from Observation.SBExecutionMode import SBExecutionMode, SBExecutionError
from Observation.Global import simulatedArray, addClassToMode, OperatorLogger
from Observation.PointingCalTarget import PointingCalTarget
from Observation.SBRatioCalTarget import SBRatioCalTarget
from Observation.TargetUtility import *
from Observation.ScanList import ScanList
from Observation.SBExecState import SBExecState
import Observation.SSRLogger



# Ugly hack to get nice logging when we're not written as a class
class StandardInterferometry(Observation.SSRLogger.SSRLogger):
    def __init__(self):
        Observation.SSRLogger.SSRLogger.__init__(self, self.__class__.__name__)
logger = StandardInterferometry()



sb = Observation.SchedulingBlock.SchedulingBlock()
logger.logInfo('Running %s scheduling block' % sb._controlSB.name)

B_max = 450.0 # length, in meters, of the longest baseline in the array to be executed.  Could either
              # get this from the resolution requirement or from the scheduler with the configuration
#
# Get expert parameters as needed here
#
numPhaseCal     = int(sb.getExpertParameter('NumPhaseCal',  default=1))
logger.logInfo('NumPhaseCal set to  %s' % numPhaseCal)
numAmplitudeCal = int(sb.getExpertParameter('NumAmplitudeCal', default=1))
logger.logInfo('NumAmplitudeCal set to  %s' % numAmplitudeCal)
numBandpassCal  = int(sb.getExpertParameter('NumBandpassCal', default=1))
numPointingCal  = int(sb.getExpertParameter('NumPointingCal', default=1))
numFocusCal     = int(sb.getExpertParameter('NumFocusCal', default=1))
numPolarizationCal = int(sb.getExpertParameter('NumPolarizationCal', default=1))
elLimit         = float(sb.getExpertParameter('ElLimit', default=20.))
maxPointingSeparation = float(sb.getExpertParameter('MaxPointingSeparation', default=25.))
sourceCycle     = float(sb.getExpertParameter('SourceCycleTime', default=0))
useScanSequence = bool(int(sb.getExpertParameter('useScanSequence', default=True)))
useDualMode     = bool(int(sb.getExpertParameter('useDualMode', default=False)))
fastSWEnabled   = bool(int(sb.getExpertParameter('FastSWEnabled', default=False)))
# Below are parameters for DiffGainCalTarget
dgcExtCycle   = float(sb.getExpertParameter('dgcExtCycle', default=600.))
dgcIntCycle   = float(sb.getExpertParameter('dgcIntCycle', default=60.))
dgcRefIntTime = float(sb.getExpertParameter('dgcRefIntTime', default=120.))
dgcSciIntTime = float(sb.getExpertParameter('dgcSciIntTime', default=180.))
dgcSourceRa   = str(sb.getExpertParameter('dgcSourceRa', default=''))  # hh:mm:ss.sss
dgcSourceDec  = str(sb.getExpertParameter('dgcSourceDec', default='')) # dd.mm.ss.sss
dgcPostpone   = float(sb.getExpertParameter('dgcPostpone', default=300.))
useDGC        = int(sb.getExpertParameter('useDGC', default=0))
#logger.logInfo('%s' % sb.getExpertParameters())

targetInfo      = None
logger.logInfo("useScanSequence value %s" % useScanSequence)
logger.logInfo('fastSWEnabled=%s' % fastSWEnabled)

if ( sourceCycle <= 0 ):
    sourceCycle = 12.0/(B_max*7.29e-5)
logger.logInfo('Source Cycle Time: %f s' % sourceCycle)
sessionControl  = int(sb.getExpertParameter('SessionControl', default=0))
logger.logInfo('Session control %d' % sessionControl)
logger.logInfo('Simulated Array %s' % simulatedArray())

# Get reference to the array and start the exec block
array = getArray()

array.beginExecution()

# Get the Observing Mode and set the maximum execution time
obsmode = array.getInterferometryObservingMode()

# add the SB execution methods
# This is a dangerous way to overwrite the parent class methods
# We should re-think how to do this (PC)
addClassToMode(obsmode, SBExecutionMode)



sbe = SBExecState(array, sb)
if ( (sessionControl) and (not simulatedArray()) ):
    # This should go into the SBExecutionMode class, but I'll do it once I have a clear idea
    # on what it has to return to the script.
    # Simple test here
    # For debugging purposes
    atSetNames   = sbe.getAttenSettingNames()
    logger.logInfo('Current list of attenuators settings %s' % atSetNames)
    logger.logInfo('Session state? %s' % sbe.getSessionState())
    try:
         if (sbe.getSessionState()):
            if ( sbe.differentSB() ):
                logger.logInfo('Current SB %s is being executed for the first time' % sb._controlSB.name)
                #logger.logInfo('Cleaning previous list of attenuator settings')
                sbe.initSession(time(), sb._controlSB.name)
                logger.logWarning('At this point we can not clear the ExecState')
                logger.logInfo('Starting first execution')
            else:
                sbe.setFirstTime(False)
                sbe.updateTime(time())
                logger.logInfo('Current SB in the ExecState is %s' % sbe.getSBName()) #exStateDict['sbName'])
                logger.logInfo('Current SB executed already %d times' % sbe.getNumberOfExecutions()) #exStateDict['sbExecuted'])
                #elapsedTime = exStateDict['sbT0'] - exStateDict['sbTimeExecuted'][-1]
                #logger.logInfo('Session has been running for %f seconds' % elapsedTime)
                logger.logInfo('Starting next execution')
         elif  (not simulatedArray()):
            logger.logInfo('Current SB %s is being executed for the first time' % sb._controlSB.name)
            sbe.initSession(time(), sb._controlSB.name)
            logger.logInfo('Starting first execution')
    except Exception, ex:
         logger.logError("Could not read the ExecState")
         logger.logError("Exception caught: %s" % ex)
         msg    = "Cannot access the ExecSate, Session failed %s" % ex
         status = ex
         raise SBExecutionError, msg

else:
    numBandpassDone = 0
    numAmplitudeDone = 0
    logger.logInfo('Not in a session, setting the session state to False')
    sbe.setSessionState(False)

# pass the sbe to the obsmode to be visible by the ifNeeded method at the target level
obsmode.addSBExecState(sbe)

# Just start on the sky, to be sure. Should be in the mode defaults.
obsmode.setCalibrationDevice(PyCalibrationDevice.NONE)

# Set observing mode restrictions
logger.logInfo("Setting elevation limit to %f degrees" % elLimit)
obsmode.setElevationLimit('%f deg' % elLimit)
endTime = sb.getMaximumExecutionTime()
logger.logInfo("Setting maximum execution time to %s seconds or %s minutes" % (endTime,endTime/60.0))
obsmode.setMaximumSchedBlockTime(endTime)
logger.logInfo("Setting max. pointing separation  to %f degrees" % maxPointingSeparation)
PointingCalTarget.setMaximumSeparation('%f deg' % maxPointingSeparation )

# Turn off automatic WVR reduction on all targets:
# we do not do this in simulation as disableWVRCalibration() is not simulated.
if ( not simulatedArray() ):
    obsmode.disableWVRCalibration(True)

# This should be the default for real interferometry: we use reference positions only for AtmCals.
for target in sb.getTargetList():
    target.setUseReferencePosition(target in sb.getAtmCalTargetList())
#
# Get group list
groupList = sb.getGroupList()

t0 = time()
#obsmode.adjustDoppler(sb.getTargetList(), sb.getScienceTargetList())
obsmode.adjustDoppler(sb.getTargetList(), sb.getRepresentativeTarget())
logger.logInfo('Doppler adjustment took %2.2f seconds' %(time()-t0))

if (len(groupList[1].getTargetList()) > 0):
    targetInfo = obsmode.getScienceTargetInfo(groupList[1])
    targetInfo['numBandpassCal'] = numBandpassCal
    targetInfo['numPhaseCal']    = numPhaseCal
    logger.logInfo("Science targets are up")
    logger.logInfo("Science targets coordinates %s" % targetInfo['coords'])
    logger.logInfo("Science targets BB frequencies %s" % targetInfo['bbfreq'])
# Check the visibility of science targets in all groups, (they should need no query expansion)
# Get the coordinates from the science target an pass it for the query
for group in groupList:
    if not obsmode.checkScienceTargetVisibility(group):
        logger.logError('Science targets in group %d are not visible, try later or tomorrow.' \
                        %(groupList.index(group)+1))
        raise SBExecutionError, "No visible science target"

# Expand source queries, then associate atm and pointing targets with other targets:
for group in groupList:
    logger.logInfo("Preparing targets for Observing Group %d" % (groupList.index(group)+1))
    t0 = time()
    #group.prepareTargets(maxPointingSeparation = maxPointingSeparation)
    group.prepareTargets(targetInfo)

# fill in Atm and SB Ratio Cal targets:
# I have to resolve the queries first!
t0 = time()
sb.createAtmCalTargets(targetInfo)
sb.createSBRatioCalTargets()
logger.logInfo('Preparation of Atm and SBRatio Cal targets took %2.2f seconds' %(time()-t0))
obsmode.adjustDoppler(sb.getAtmCalTargetList()+sb.getSBRatioCalTargetList(), sb.getRepresentativeTarget())


logger.logInfo("")
logger.logInfo("")
logger.logInfo("This is a test, listing all targets and sources")
for i, target in enumerate(sb.getTargetList()):
    logger.logInfo("Target %s number  %d  with source %s" % (target, i, target.getSource().sourceName ))
logger.logInfo("")
logger.logInfo("")

sb.setTargetsId()

#Recycle the Field Sources
logger.logInfo('In a session? %s' % sbe.getSessionState())
if ( sessionControl and (sbe.getSessionState()) ):
    logger.logInfo("Now we are recycling field sources from the Exec State")
    obsmode.recycleFieldSources(sbe, sb)
#    obsmode.moveCalsFromGroupOne(groupList)

#Check whether to recycle the attenuators
if ( (sessionControl) and (sbe.getSessionState()) ):
     # This method has to decide whether I will reuse the setting or force a new optimization
     # Do I need default values for all dictionary entries? As it is passed as a string 
     # it can always be re-created.
    logger.logInfo("Now we are recycling attenuator settings")
    obsmode.recycleAttenSettings(sbe)



# None means OK. If an error occurs the status is replaced, in the catch block, by the exception.
status = None
    
try:
    # The first group is for SB calibrators, we don't expect any science target.
    group = groupList[0]
    groupNumber = groupList.index(group)+1
    logger.logInfo("Start Observing Group %d (Calibrations)" % groupNumber)
    #
    # Pointing now executed in association with the other cal targets.
    if (useScanSequence):
        logger.logInfo("Scan Sequences are enabled")
        scanList = ScanList()
    else:
        scanList = None
        logger.logInfo("Scan Sequences are disabled")

### Adding an option to add dual mode...
    if useDualMode :
        for target in sb.getTargetList() :
            ss = target._spectralSpec
            if not hasattr(ss,'SquareLawSetup') :
                ss.SquareLawSetup = CCL.APDMSchedBlock.SquareLawSetup()
                ss.SquareLawSetup.integrationDuration.set(0.1)
 
    nFocus, nFocusCleanup = obsmode.executeCalTargetList(group.getFocusCalTargetList(), minimumNumber=numFocusCal)
    
    #
    #Begin merge
    #
    # The "if" below is for sessions, not important now
    #if numBandpassDone < numBandpassCal:

    # Polarization section needs to be cleanup a bit.
    numDone, numInCleanup = obsmode.executeCalTargetList(group.getPolarizationCalTargetList(),
                                                         minimumNumber = numPolarizationCal,
                                                         deferToCleanupList=True,scanList=scanList)

    if (useScanSequence):
        scanList.execute(obsmode)
    if group.getPolarizationCalTargetList() is not () :
        obsmode.addTargetToCleanupList(group.getPolarizationCalTargetList()[0])

    logger.logInfo("Polarization Targets: %d executed, %d in Cleanup List" % (numDone, numInCleanup))

    numDone, numInCleanup = obsmode.executeCalTargetListSpec(group.getBandpassCalTargetList(),
                                                             minimumNumber=numBandpassCal,
                                                             deferToCleanupList=True,scanList=scanList)

    if (useScanSequence):
       scanList.execute(obsmode)
    logger.logInfo("Bandpass Targets: %d executed, %d in Cleanup List" % (numDone, numInCleanup))

    #
    #End merge
    #

    obsmode.executeCalTargetList(group.getDelayCalTargetList(),scanList=scanList)
    if (useScanSequence):
        scanList.execute(obsmode)


    #
    #Begin merge
    #
    logger.logInfo(" ")
    logger.logInfo("Starting Amplitude Cal Target execution" )
    numDone, numInCleanup = obsmode.executeCalTargetList(group.getAmplitudeCalTargetList(),
                                            minimumNumber=numAmplitudeCal,
                                            deferToCleanupList=True,scanList=scanList)
    logger.logInfo("Amplitude Cal Targets: %d executed, %d in Cleanup List" \
                 %(numDone, numInCleanup))

    if (useScanSequence):
        scanList.execute(obsmode)
    #logger.logInfo("Amplitude Cal Targets: %d executed, %d in Cleanup List" %(numDone, numInCleanup))
    #
    #End merge
    #

    # Now the rest of the groups should contain science targets.
    for group in groupList[1:]:
        groupNumber = groupList.index(group)+1
        # The primary phase calibrator is assumed to be the one with the
        # shortest cycle time in the group.  This effectivly sets the scan
        # duration.
        primaryPhaseCal = obsmode.getPrimaryPhaseCal(group)
        if primaryPhaseCal is None:
            logger.logError("No phase calibrator available in Observing Group %s" % groupNumber)
            raise SBExecutionError, "No phase calibrator"
        else:
            logger.logInfo("Selected %s as Primary phase calibrator" % primaryPhaseCal)
        logger.logInfo("Start Observing Group %d (Science)" % groupNumber)

        # construct DiffGainCalTarget, if needed
        if useDGC:
            kwargs = dict(
                dgcExtCycle=dgcExtCycle,
                dgcIntCycle=dgcIntCycle,
                dgcRefIntTime=dgcRefIntTime,
                dgcSciIntTime=dgcSciIntTime,
                dgcSourceRa=dgcSourceRa,
                dgcSourceDec=dgcSourceDec,
                dgcPostpone=dgcPostpone,
            )
            isB2B = group.isBandToBand()
            logger.logInfo('isB2B = %s' % (isB2B))
            # Both B2B and BW switching observations can construct DGC in a same way.
            diffGainCalTarget = group.constructDiffGainCalTargetForB2B(obsmode, **kwargs)

        #
        # A group is defined to be complete when all science targets within
        # the group are complete (or unobservable)
        obsmode.assignSubcycleTime(group.getScienceTargetList(), sourceCycle)
        while not group.isComplete(obsmode):
            outerCycleTargets = list(group.getFocusCalTargetList()) + \
                                list(group.getAmplitudeCalTargetList()) + \
                                list(group.getPolarizationCalTargetList()) + \
                                list(group.getBandpassCalTargetList()) + \
                                list(group.getDelayCalTargetList())
            if useDGC:
                outerCycleTargets.append(diffGainCalTarget)

            secondaryPhaseCals = list(group.getPhaseCalTargetList())
            secondaryPhaseCals.remove(primaryPhaseCal)

            obsmode.executeCalTargetList(group.getFocusCalTargetList(),scanList=scanList)
            if (useScanSequence):
                scanList.execute(obsmode)

            obsmode.executeCalTargetList(group.getPolarizationCalTargetList(),minimumNumber = numPolarizationCal, scanList=scanList)
            if (useScanSequence):
                scanList.execute(obsmode)

            obsmode.executeCalTargetList(group.getBandpassCalTargetList(),scanList=scanList)
            if (useScanSequence):
                scanList.execute(obsmode)

            obsmode.executeCalTargetList(group.getAmplitudeCalTargetList(),scanList=scanList)
            if (useScanSequence):
                scanList.execute(obsmode)

            if useDGC:
                # DiffGainCal
                # For B2B observations, it is desired to execute DGC just before the primary phase cal execution.
                isExecuted = diffGainCalTarget.executeIfNeeded(obsmode, scanList=scanList)
                if isExecuted:
                    # Force execution of next primary phase calibrator.
                    primaryPhaseCal.overwriteNextExecutionTime(value=0.)

            obsmode.executeCalTargetList(group.getPhaseCalTargetList(),
                                         minimumNumber=numPhaseCal,
                                         deferToCleanupList=True,
                                         minimumInCleanupList=numPhaseCal,
                                         scanList=scanList)
            if not fastSWEnabled and useScanSequence:
                scanList.execute(obsmode)

            # The DelayCal here is a substitute for a Test Source.
            # It is executed at every outer loop execution on the target.
            obsmode.executeCalTargetList(group.getDelayCalTargetList(),scanList=scanList)
            if not fastSWEnabled and useScanSequence:
                scanList.execute(obsmode)

            while not group.isComplete(obsmode):
                # Inner-cycle: if scan-sequence is enabled, try to sequence as many as possible science and phase scans.
                # execute science targets until either a calibration is needed
                obsmode.executeScienceTargetList(group, outerCycleTargets + [primaryPhaseCal], scanList=scanList)
                if not fastSWEnabled and useScanSequence:
                    scanList.execute(obsmode)
                # then, phase-cal
                primaryPhaseCal.execute(obsmode, scanList=scanList)

                if not fastSWEnabled and useScanSequence:
                    scanList.execute(obsmode)

                if not primaryPhaseCal.isObservable(obsmode, duration=primaryPhaseCal.getCycleTime(), scanList=scanList):
                    primaryPhaseCal = obsmode.getPrimaryPhaseCal(group, primaryPhaseCal)
                    if primaryPhaseCal is None:
                        raise Exception('No primary phase calilbrator is available')

                if not fastSWEnabled:
                    # If fast-sw disabled, do not try to sequence phase and science targets
                    break
                if (useScanSequence) and scanList.getCurrentCompletionTime() > 600:
                    break

                # loop-end condition. (If outer-cycle targets need execution, get out of inner loop)
                currentElapsedTime = obsmode.sbe.getElapsedTime() if (sessionControl and sbe.getSessionState()) else obsmode.getElapsedTime()
                if useScanSequence:
                    currentElapsedTime += scanList.getCurrentCompletionTime()
                nextReqTimes = [target.getNextRequiredTime() for target in outerCycleTargets + secondaryPhaseCals]
                if len(nextReqTimes) > 0 and min(nextReqTimes) <= currentElapsedTime:
                    break

            # Exexcute sequence of science and primary phase cal scans
            if (useScanSequence):
                scanList.execute(obsmode)

        # Group is complete: remove the primaryPhaseCal from cleanup:
        logger.logInfo("Observing Group %s is completed."% groupNumber)
        
        obsmode.removeTargetFromCleanupList(primaryPhaseCal)

    # All groups are complete: execute the cleanup (e.g. bandpass or amplitude cals)
    logger.logInfo("All groups are completed.")
    logger.logInfo("Executing cleanup list.")
    obsmode.executeCleanupList()
    
except SBOutOfTime:
    #
    # Ok there should be just enough time for the cleanup list now
    groupNumber = groupList.index(group)
    # In principle sessionControl is not needed here
    if ( (sessionControl) and (groupNumber > 0) and sbe.getSessionState() ):
       currentIndex = sbe.getScienceTargetIndex()
       if (currentIndex == group._subcycleIndex):
          group._subcycleIndex += 1
          group._subcycleIndex %= len(list(group.getScienceTargetList()))
       sbe.addScienceTargetIndex(group._subcycleIndex)
       logger.logInfo('Moving Index %d into the ExecState' %  group._subcycleIndex)
    if (useScanSequence):
        scanList.execute(obsmode)
    logger.logWarning("Running short of time, forcing cleanup!")
    logger.logInfo("Executing cleanup list.")
    obsmode._maximumTime = None
    obsmode.executeCleanupList()

except Exception, ex:
    logger.logError("Exception caught: %s" % ex)
    status = ex
    raise
    
finally:
    if (sessionControl) and sbe.getSessionState():
       try:
          tf = time()
          logger.logInfo("Saving SB State in the ExecState at timestamp %f" % tf)
       except:
          tf = 1.0
          logger.logInfo("Saving SB State in the ExecState time() called failed")
       if (not simulatedArray()):
          sbe.endSBExecution(tf)

    try:
        obsmode.resetLimits()
    except:
        logger.logError("Could not reset limits")

    array.endExecution(status)

#
# ___oOo___
